import typing
import numpy

class WavefrontObj:

    def __init__(self):
        self.vtxxyz2xyz = None
        self.vtxuv2uv = None
        self.vtxnrm2nrm = None
        self.elem2idx = None
        self.idx2vtxxyz = None
        self.idx2vtxuv = None
        self.idx2vtxnrm = None
        self.elem2group = None
        self.group2name = None
        self.elem2mtl = None
        self.mtl2name = None
        self.mtl_file_name = None

    def centerize(self):
        self.vtxxyz2xyz[:] -= (self.vtxxyz2xyz.max(axis=0) + self.vtxxyz2xyz.min(axis=0)) * 0.5

    def normalize_size(self, scale=1.0):
        self.vtxxyz2xyz *= scale / (self.vtxxyz2xyz.max(axis=0) - self.vtxxyz2xyz.min(axis=0)).max()

    def edges(self):
        from .del_msh import edges_of_polygon_mesh
        return edges_of_polygon_mesh(self.elem2idx, self.idx2vtxxyz, self.vtxxyz2xyz.shape[0])

    def tris(self):
        from .del_msh import triangles_from_polygon_mesh
        return triangles_from_polygon_mesh(self.elem2idx, self.idx2vtxxyz)

    def triangle_mesh_with_uv(self):
        from .del_msh import triangles_from_polygon_mesh, unify_two_indices_of_triangle_mesh

        tri2vtxxyz = triangles_from_polygon_mesh(self.elem2idx, self.idx2vtxxyz)
        tri2vtxuv = triangles_from_polygon_mesh(self.elem2idx, self.idx2vtxuv)

        tri2uni, uni2vtxxyz, uni2vtxuv = unify_two_indices_of_triangle_mesh(tri2vtxxyz, tri2vtxuv)

        uni2xyz = numpy.ndarray((uni2vtxxyz.shape[0], 3), self.vtxxyz2xyz.dtype)
        uni2xyz[:, :] = self.vtxxyz2xyz[uni2vtxxyz[:], :]

        uni2uv = numpy.ndarray((uni2vtxuv.shape[0], 2), self.vtxuv2uv.dtype)
        uni2uv[:, :] = self.vtxuv2uv[uni2vtxuv[:], :]
        return tri2uni, uni2xyz, uni2uv, uni2vtxxyz, uni2vtxuv

    def polygon_mesh_with_normal(self):
        from .del_msh import unify_two_indices_of_polygon_mesh
        idx2uni, uni2vtxxyz, uni2vtxnrm = unify_two_indices_of_polygon_mesh(
            self.elem2idx, self.idx2vtxxyz, self.idx2vtxnrm)

        uni2xyz = numpy.ndarray((uni2vtxxyz.shape[0], 3), self.vtxxyz2xyz.dtype)
        uni2xyz[:, :] = self.vtxxyz2xyz[uni2vtxxyz[:], :]

        uni2nrm = numpy.ndarray((uni2vtxnrm.shape[0], 3), self.vtxnrm2nrm.dtype)
        uni2nrm[:, :] = self.vtxnrm2nrm[uni2vtxnrm[:], :]
        return idx2uni, uni2xyz, uni2nrm, uni2vtxxyz, uni2vtxnrm


    def extract_polygon_mesh_of_material(self, mtl_idx):
        from .del_msh import extract_flagged_polygonal_element
        return extract_flagged_polygonal_element(
            self.elem2idx, self.idx2vtxxyz, self.elem2mtl[:] == mtl_idx)


def load(file_path: str, is_centerize = False, normalized_size: typing.Optional[float]=None):
    from .del_msh import load_wavefront_obj
    o = WavefrontObj()
    o.vtxxyz2xyz, o.vtxuv2uv, o.vtxnrm2nrm, \
    o.elem2idx, o.idx2vtxxyz, o.idx2vtxuv, o.idx2vtxnrm, \
    o.elem2group, o.group2name, \
    o.elem2mtl, o.mtl2name, o.mtl_file_name = load_wavefront_obj(file_path)
    if is_centerize:
        o.centerize()
    if type(normalized_size) == float:
        o.normalize_size(normalized_size)
    return o


def read_material(path: str):
    with open(path) as f:
        dict_mtl = {}
        cur_mtl = {}
        cur_name = ""
        for line in f:
            if line.startswith('#'):
                continue
            words = line.split()
            if len(words) == 2 and words[0] == 'newmtl':
                cur_name = words[1]
                cur_mtl = {}
            if len(words) == 0 and cur_name != "":
                dict_mtl[cur_name] = cur_mtl
            if len(words) == 4 and words[0] == 'Kd':
                cur_mtl['Kd'] = (float(words[1]), float(words[2]), float(words[3]))
    return dict_mtl



def triangles(pelem2pidx, pidx2vtxxyz):
    from .del_msh import triangles_from_polygon_mesh
    return triangles_from_polygon_mesh(pelem2pidx, pidx2vtxxyz)


def extract(elem2idx, idx2vtx, elem2bool):
    from .del_msh import extract_flagged_polygonal_element
    return extract_flagged_polygonal_element(elem2idx, idx2vtx, elem2bool)


def edges(elem2idx, idx2vtx, num_vtx):
    from .del_msh import edges_of_polygon_mesh
    return edges_of_polygon_mesh(elem2idx, idx2vtx, num_vtx)



import typing
import numpy.typing


def edge2vtx(tri2vtx: numpy.typing.ArrayLike, num_vtx: int):
    assert len(tri2vtx.shape) == 2
    assert tri2vtx.shape[1] == 3
    from .del_msh import edges_of_uniform_mesh
    return edges_of_uniform_mesh(tri2vtx, num_vtx)


def vtx2vtx(tri2vtx, num_vtx: int):
    assert len(tri2vtx.shape) == 2
    assert tri2vtx.shape[1] == 3
    from .del_msh import vtx2vtx_trimesh
    return vtx2vtx_trimesh(tri2vtx, num_vtx)


def triangle_adjacency(tri2vtx, num_vtx: int):
    from .del_msh import elsuel_uniform_mesh_polygon
    return elsuel_uniform_mesh_polygon(tri2vtx, num_vtx)


def topological_distance_of_tris(idx_tri, tri2tri):
    from .del_msh import topological_distance_on_uniform_mesh
    return topological_distance_on_uniform_mesh(idx_tri, tri2tri)

# above: topology
# ------------------------
# below: primitive

def torus(major_radius=1.0, minor_radius=0.4, ndiv_major_radius=32, ndiv_minor_radius=32):
    from .del_msh import torus_meshtri3
    return torus_meshtri3(major_radius, minor_radius, ndiv_major_radius, ndiv_minor_radius)


def capsule(radius=1.0, height=1., ndiv_theta=32, ndiv_height=32, ndiv_longtitude=32):
    from .del_msh import capsule_meshtri3
    return capsule_meshtri3(radius, height, ndiv_theta, ndiv_longtitude, ndiv_height)


def cylinder(radius=1., height=1., ndiv_theta=32, ndiv_height=8):
    from .del_msh import cylinder_closed_end_meshtri3
    return cylinder_closed_end_meshtri3(radius, height, ndiv_theta, ndiv_height)


def sphere(radius=1., ndiv_latitude=32, ndiv_longtitude=32):
    from .del_msh import sphere_meshtri3
    return sphere_meshtri3(radius, ndiv_latitude, ndiv_longtitude)


def load_wavefront_obj(
        path_file: str,
        is_centerize=False,
        normalized_size: typing.Optional[float] = None):
    from .del_msh import load_wavefront_obj_as_triangle_mesh
    tri2vtx, vtx2xyz = load_wavefront_obj_as_triangle_mesh(path_file)
    if is_centerize:
        vtx2xyz[:] -= (vtx2xyz.max(axis=0) + vtx2xyz.min(axis=0)) * 0.5
    if type(normalized_size) == float:
        vtx2xyz *= normalized_size / (vtx2xyz.max(axis=0) - vtx2xyz.min(axis=0)).max()
    return tri2vtx, vtx2xyz


def unindexing(tri2vtx, vtx2xyz):
    from .del_msh import unidex_vertex_attribute_for_triangle_mesh
    return unidex_vertex_attribute_for_triangle_mesh(tri2vtx, vtx2xyz)


def areas(tri2vtx, vtx2xyz):
    from .del_msh import areas_of_triangles_of_mesh
    return areas_of_triangles_of_mesh(tri2vtx, vtx2xyz)


# above: property
# ---------------------
# below: search

def first_intersection_ray(src, dir, tri2vtx, vtx2xyz):
    from .del_msh import first_intersection_ray_meshtri3
    return first_intersection_ray_meshtri3(src, dir, tri2vtx, vtx2xyz)


def pick_vertex(src, dir, tri2vtx, vtx2xyz):
    from .del_msh import pick_vertex_meshtri3
    return pick_vertex_meshtri3(src, dir, tri2vtx, vtx2xyz)


# above: search
# ------------------------
# below: sampling

def position(
        tri2vtx: numpy.typing.NDArray, vtx2xyz: numpy.typing.NDArray,
        idx_tri: int, r0: float, r1: float):
    i0 = tri2vtx[idx_tri][0]
    i1 = tri2vtx[idx_tri][1]
    i2 = tri2vtx[idx_tri][2]
    p0 = vtx2xyz[i0]
    p1 = vtx2xyz[i1]
    p2 = vtx2xyz[i2]
    return r0 * p0 + r1 * p1 + (1. - r0 - r1) * p2


def sample(cumsum_area: numpy.typing.NDArray, r0: float, r1: float):
    from .del_msh import sample_uniformly_trimesh
    return sample_uniformly_trimesh(cumsum_area, r0, r1)


def sample_many(tri2vtx, vtx2xy, num_sample: int):
    import random
    tri2area = areas(tri2vtx, vtx2xy)
    cumsum_area = numpy.cumsum(numpy.append(0., tri2area)).astype(numpy.float32)
    xys = []
    for i in range(num_sample):
        smpl_i = sample(cumsum_area, random.random(), random.random())
        pos_i = position(tri2vtx, vtx2xy, *smpl_i)
        xys.append(pos_i.tolist())
    return numpy.array(xys)


###################################

def merge_hessian_mesh_laplacian(
        tri2vtx, vtx2xyz,
        row2idx,
        idx2col,
        row2val, idx2val):
    from .del_msh import merge_hessian_mesh_laplacian_on_trimesh
    merge_hessian_mesh_laplacian_on_trimesh(
        tri2vtx, vtx2xyz,
        row2idx, idx2col,
        row2val, idx2val)
